class Model:
    search_spaces = {}
    system_params = {}
    system_fit_params = {}
    Estimator = None
    short_name = 'model'
    
    def __init__(self, params=None, n_jobs=-1, verbosity=False):
        """
        :param params: model hyperparameters
        :param n_jobs: number of threads to use
        :param verbosity: True or False (print logs or not)
        """
        self.is_fit = 0
        self.params = params
        self.n_jobs = n_jobs
        self.verbosity = verbosity
        if self.params:
            self.estimator = type(self).Estimator(**self.params, **type(self).system_params)
        else:
            self.estimator = type(self).Estimator(**type(self).system_params)
        self.reset_name()
        
    def reset_name(self):
        """
        Changes name according to current parameters.
        :return:
        """
        self.__name__ = f"{type(self).short_name}_{hex(hash(frozenset(self.params.items())))[-2:] if self.params else 'default'}"
        
    def fit(self, X, y, **kwargs):
        """
        Fit model. You can call this method only once
        :param X: data matrix
        :param y: target
        :param kwargs: additional params to be used for model fitting
        :return:
        """
        if self.is_fit:
            raise UserWarning(f"This {type(self).__name__} is already fit.\nUse .warm_fit() to fit it from current state.")
        fit_params = {}
        for param in type(self).system_fit_params:
            if param in kwargs:
                fit_params[param] = kwargs[param]
            else:
                fit_params[param] = type(self).system_fit_params[param]
        self._fit(X, y, **fit_params)
        self.is_fit = True

    def _fit(self, X, y, **kwargs):
        """
        This method is one of which you need to override to define a custom model.
        :param X: data matrix
        :param y: target values
        :param kwargs: additional params
        :return:
        """
        self.estimator.fit(X, y, **kwargs)
        
    def __setattr__(self, name, value):
        if name == 'params':
            if self.is_fit:
                raise UserWarning(f"Can't change params of trained model")
            else:
                super().__setattr__(name, value)
                if not self.params:
                    return
                self.estimator = type(self).Estimator(**self.params, **type(self).system_params)
                self.reset_name()
        else:
            super().__setattr__(name, value)
        
    def predict(self, X):
        """
        Standard prediction method.
        :param X: data matrix
        :return:
        """
        return self.estimator.predict(X)
    
    def __mul__(self, x):
        return WeightedModel(self, x)
    
    def __rmul__(self, x):
        return self * x
    
    def __truediv__(self, x):
        assert x != 0
        return WeightedModel(self, 1. / x)
    
    def __add__(self, other):
        assert type(self).__base__ == type(other).__base__, \
        f"Can't add {type(other).__name__} to {type(self).__name__}."
        return Ensemble([self, other])
    
    def __radd__(self, other):
        return self + other
        
    def __repr__(self):
        return self.__name__
        

class WeightedModel(Model): # MulNode
    """
    Multiplies predictions by a certain coefficient.
    """
    def __init__(self, model, coeff):
        self.model = model
        self.coeff = coeff
        if type(self.model) == type(self):
            self.coeff *= self.model.coeff
            self.model = self.model.model
        self.__name__ = f"{round(self.coeff, 2)} * "
        if '+' in self.model.__name__:
            self.__name__ += f"({self.model.__name__})"
        else:
            self.__name__ += f"{self.model.__name__}"

    def predict(self, X):
        """
        Standard prediction method.
        :param X: data matrix
        :return:
        """
        return self.coeff * self.model.predict(X)
    

class Ensemble(Model): # AddNode
    """
    Sums up all predictions of all models.
    """
    def __init__(self, models):
        self.models = []
        ensembles = [model for model in models if type(model) == type(self)]
        single_models = [model for model in models if type(model) != type(self)]
        _models = []
        for ensemble in ensembles:
            _models += ensemble.models
        _models += single_models
        
        def __get_model_coeff(model):
            """
            :param model:
            :return: if the model is weighted, then returns its coefficient, else 1
            """
            if isinstance(model, WeightedModel):
                return model.model, model.coeff
            return model, 1
        
        self.models = list(set([__get_model_coeff(model)[0] for model in _models]))
        _coeffs = [0 for i in range(len(self.models))]
        for i in range(len(self.models)):
            for model in _models:
                if __get_model_coeff(self.models[i])[0] == __get_model_coeff(model)[0]:
                    _coeffs[i] += __get_model_coeff(model)[1]
        
        self.models = [WeightedModel(model, coeff) for model, coeff in zip(self.models, _coeffs)]
                    
        self.__name__ = ' + '.join([model.__name__ for model in self.models])
        
    def predict(self, X):
        """
        Standard prediction method.
        :param X: data matrix
        :return:
        """
        res = 0
        for model in self.models:
            res += model.predict(X)
        return res
    
    def __mul__(self, x):
        return Ensemble([model * x for model in self.models])
    
    def __truediv__(self, x):
        assert x != 0
        return Ensemble([model / x for model in self.models])


from ..storage import source_utils
from ..storage.caching import cache
from .. import config
from .storage import FeatureConstructor
from ..storage import dataframe
from IPython.display import display
from glob import glob
import os


def preview(df, sizes=(2, 4, 6)):
    """
    Applies function to heads of particular dataframe.
    
    Example:
    ``` python
    @preview(df, sizes=[5, 15])
    def make_ohe_pclass(df):
        ...
    ```
    """

    def __preview(function):
        config.preview_call = 1
        for sz in sizes:
            # if isinstance(df, dataframe.DataFrame):
            #     ktdf = dataframe.DataFrame(df.head(sz), df.train, df.encoders)
            # else:
            #     ktdf = dataframe.DataFrame(df.head(sz), True, {})
            ktdf = dataframe.DataFrame(df.head(sz), True, {})  # not a bug, but a feature: here we use KTDF's property that its methods return pd.DataFrames
            display(function(ktdf))
        config.preview_call = 0

    return __preview


def register(*args, cache_default=True):
    """
    Registers function for further caching its calls and restoring source.
    
    Example:
    ``` python
    @register
    def make_ohe_pclass(df):
        ...
    ```
    """

    def __register(func):
        if source_utils.source_is_saved(func) and not source_utils.matches_cache(func):
            raise NameError("A function with the same name is already registered")

        functor = FeatureConstructor(func, cache_default)
        if not source_utils.source_is_saved(func):
            cache.cache_obj(functor, functor.__name__ + '_fc')
            source_utils.save_source(func)
        return functor

    if args:
        function = args[0]
        return __register(function)
    else:
        return __register


def deregister(name, force=False):
    """
    Deletes sources and cached calls of a certain function.
    The interface is too rich now:

    ``` python
    @deregister
    def make_new_features(df):
        ...

    @deregister(force=False)
    def make_new_features(df):
        ...

    deregister('make_new_features', force=False)

    @deregister(force=True)
    def make_new_features(df):
        ...

    deregister('make_new_features', force=True)
    ```

    It's highly recommended to use only `deregister('function_name')` interface.
    Other ones are deprecated.
    """
    confirmation = ''
    if not force:
        print("Are you sure you want to delete all the cached files?")
        print("To confirm please print full name of the function:")
        confirmation = input()
    if not force and confirmation != name:
        print("Doesn't match")
        return

    paths = glob(config.storage_path + name + '_fc_obj') + \
            glob(config.storage_path + name + '__[0-9a-f][0-9a-f][0-9a-f][0-9a-f]_df') + \
            glob(config.source_path + name + '.py')
    for path in paths:
        print(f'removing {path}')
        os.remove(path)


def dropper(function):
    """
    Registers function that won't be cached.
    Is recommended to be used only with functions which actually drop columns or rows and don't produce any new data.
    
    Example:
    ``` python
    @dropper
    def drop_pclass(df):
        return stl.column_dropper(['Pclass'])(df)
    ```
    """
    #     TODO:
    #     if cache.is_cached(function.__name__):
    #         print('Dropper is already registered. Deregistering: ')
    #         deregister(function.__name__, force=True)
    deregister(function.__name__, force=True)
    return register(function, cache_default=False)


def selector(function):
    """
    Registers function that won't be cached.
    Is recommended to be used only with functions which actually select columns or rows and don't produce any new data.
    
    Example:
    ``` python
    @selector
    def select_pclass_cabin(df):
        return stl.column_selector(['Pclass', 'Cabin'])(df)
    ```
    """

    deregister(function.__name__, force=True)
    return register(function, cache_default=False)


# def trainable(function):
#     def __trainable(df, )


from .storage import FeatureConstructor
import pandas as pd
import numpy as np
from ..storage.dataframe import DataFrame as KTDF


def empty_like(df):
    return df[[]].copy()


identity = FeatureConstructor(lambda df: df, cache_default=False)
identity.stl = True
identity.source = "stl.identity"


def merge(dfs):
    if len(dfs) == 1:
        return dfs[0]
    return pd.concat([df.df if isinstance(df, KTDF) else df for df in dfs], axis=1)


def column_selector(columns):
    def __col_selector(df):
        return df[columns]

    return FeatureConstructor(__col_selector, cache_default=False)


def column_dropper(columns):
    def __col_dropper(df):
        return df.drop(columns, axis=1)

    return FeatureConstructor(__col_dropper, cache_default=False)


def concat(funcs):
    def __concat(df):
        return merge([func(df) for func in funcs])

    fc = FeatureConstructor(__concat, cache_default=False)
    fc.source = f"stl.concat([{', '.join([func.__name__ if not func.stl else func.source for func in funcs])}])"
    fc.__name__ = f"concat_{''.join([func.__name__[:2] if not func.stl else func.__name__ for func in funcs])}"
    fc.stl = True
    return fc


def compose(funcs):
    def __compose(df):
        res = empty_like(df)
        for func in funcs:
            res = merge([res, func(res)])
        return res

    fc = FeatureConstructor(__compose, cache_default=False)
    fc.source = f"stl.compose([{', '.join([func.__name__ if not func.stl else func.source for func in funcs])}])"
    fc.__name__ = f"compose_{''.join([func.__name__[:2] if not func.stl else func.__name__ for func in funcs])}"
    fc.stl = True
    return fc


# TODO: reimplement using sklearn.preprocessing.OHE
def make_ohe(cols, sep='_ohe_'):
    def __make_ohe(df):
        # encoder_name = '__ohe_' + '-'.join(cols)
        # if df.train:
        #     res = pd.get_dummies(df.df[cols], prefix_sep=sep, columns=cols)
        #     df.encoders[encoder_name] = list(res.columns)
        # else:
        #     res = pd.get_dummies(df.df[cols], prefix_sep=sep, columns=cols)
        #     for col in (set(res.columns) - set(df.encoders[encoder_name])):
        #         res[col] = 0
        #     return res[df.encoders[encoder_name]]

        return pd.get_dummies(df.df[cols], prefix_sep=sep, columns=cols)

    return FeatureConstructor(__make_ohe, cache_default=False)


def make_mean_encoding(cols, target_col, prefix='me_'):
    def __make_mean_encoding(df):
        # print("ME call: ", type(df))
        # print(df.train)
        res = empty_like(df)
        for col in cols:
            res[prefix + col] = 0
            if df.train:
                enc = dict()
                for value in set(df[col]):
                    idxs = (df[col] == value)
                    enc[value] = res.loc[idxs, prefix + col] = df[idxs][target_col].mean()
                df.encoders['__me_' + col] = enc
            else:
                enc = df.encoders['__me_' + col]
                for value in set(df[col]):
                    if value in enc.keys():
                        encoding_value = enc[value]
                    else:
                        encoding_value = np.mean(list(enc.values()))
                        # print(f'Unknown value: {value}, inplacing with {encoding_value}')
                    idxs = (df[col] == value)
                    res.loc[idxs, prefix + col] = encoding_value
        return res

    return FeatureConstructor(__make_mean_encoding, cache_default=False)


import multiprocessing
import swifter


def apply(dataframe, function, **kwargs):
    """
    Applies function to dataframe faster.
    If n_threads is in kwargs and is greater than 1, applies by multiprocessing.
    :return: same as df.apply(function)
    """
    if 'n_threads' in kwargs:
        n_threads = kwargs.pop('n_threads')
    else:
        n_threads = 1
    if n_threads == 1:
        return dataframe.swifter.apply(function, **kwargs)

    def _apply_df(args):
        df, func, num, kw = args
        return num, df.swifter.apply(func, **kw)

    pool = multiprocessing.Pool(processes=n_threads)
    result = pool.map(_apply_df, [(d, function, i, kwargs) for i, d in enumerate(np.array_split(dataframe, n_threads))])
    pool.close()
    result = sorted(result, key=lambda x: x[0])
    return pd.concat([i[1] for i in result])